/*
eis * Copyright (C) 2008,2009  OMRON SOFTWARE Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.

 * test message
 */

package net.gorry.android.input.nicownng.JAJP;

import java.util.HashMap;
import java.util.List;
import java.util.Locale;

import net.gorry.android.input.nicownng.ComposingText;
import net.gorry.android.input.nicownng.DefaultSoftKeyboard;
import net.gorry.android.input.nicownng.MyHeightKeyboard;
import net.gorry.android.input.nicownng.NicoWnnG;
import net.gorry.android.input.nicownng.NicoWnnGEvent;
import net.gorry.android.input.nicownng.NicoWnnGJAJP;
import net.gorry.android.input.nicownng.R;
import android.content.SharedPreferences;
import android.inputmethodservice.Keyboard;
import android.preference.PreferenceManager;
import android.text.InputType;
import android.util.Log;
import android.view.KeyEvent;
import android.view.View;
import android.view.inputmethod.EditorInfo;
import android.view.inputmethod.InputConnection;

/**
 * The default Software Keyboard class for Japanese IME.
 *
 * @author Copyright (C) 2009 OMRON SOFTWARE CO., LTD.  All Rights Reserved.
 */
public class DefaultSoftKeyboardNico2 extends DefaultSoftKeyboard {
	/** new keycode **/
	private static final int KEYCODE_NEWKEY_0 = -410;
	private static final int KEYCODE_NEWKEY_1 = -411;
	private static final int KEYCODE_NEWKEY_2 = -412;
	private static final int KEYCODE_NEWKEY_3 = -413;
	private static final int KEYCODE_NEWKEY_4 = -414;

	/** Input mode toggle cycle table */
	private static final int[] JP_MODE_CYCLE_TABLE = {
		KEYMODE_JA_FULL_NICO,
	};

	/* */
	SetupKeyboard              mSetupKeyboard;
	String[][][][]             mCycleTable;
	// HashMap<String, String>    mReplaceTable;

	/** Default constructor */
	public DefaultSoftKeyboardNico2(final NicoWnnG parent, final SetupKeyboard keyboard) {
		super(parent);
		mSetupKeyboard       = keyboard;
		mCycleTable   = keyboard.SetupCycleTable();

		mCurrentLanguage     = LANG_JA;
		mShiftOn             = KEYBOARD_SHIFT_OFF;
		mCurrentKeyMode      = KEYMODE_JA_FULL_NICO;
		mCurrentSlide        = NICO_SLIDE_MODE_TOP;
	}

	/** @see com.hiroshica.android.input.nicownn2.DefaultSoftKeyboard#createKeyboards */
	@Override protected void createKeyboards() {
		super.createKeyboards();

		if (m12keyTable[mCurrentKeyMode] == KEYBOARD_12KEY) {
			mWnn.onEvent(mWnn.mEventChangeMode12Key);
			NicoWnnGJAJP.getInstance().setSpecialCandidateOnKana12KeyToggleMode((mFlickNicoInput > 0) ? 2 : 0);
		} else {
			mWnn.onEvent(mWnn.mEventChangeModeQwerty);
			NicoWnnGJAJP.getInstance().setSpecialCandidateOnKana12KeyToggleMode(0);
		}
	}

	/**
	 *
	 */
	@Override public void checkHiddenKeyboard() {
		super.checkHiddenKeyboard();
	}

	/**
	 * Commit the pre-edit string for committing operation that is not explicit
	 * (ex. when a candidate is selected)
	 */
	private void commitText() {
		if (!mNoInput) {
			mWnn.onEvent(mWnn.mEventCommitText);
		}
	}

	/**
	 * Change input mode
	 * <br>
	 * @param keyMode   The type of input mode
	 */
	@Override
	public void changeKeyMode(final int keyMode) {
		final int targetMode = filterKeyMode(keyMode);
		if (targetMode == INVALID_KEYMODE) {
			return;
		}

		commitText();

		if (mCapsLock) {
			mWnn.onEvent(mWnn.mEventInputShiftLeft);
			mCapsLock = false;
		}
		mShiftOn = KEYBOARD_SHIFT_OFF;
		final MyHeightKeyboard kbd = getModeChangeKeyboard(targetMode);
		mCurrentKeyMode = targetMode;
		mPrevInputKeyCode = 0;
		mCurrentSlide = 0;
		mNicoKeyMode = NICO_MODE_FULL_HIRAGANA;

		int mode = NicoWnnGEvent.Mode.DIRECT;

		switch (targetMode) {

			case KEYMODE_JA_FULL_NICO:
				mInputType = INPUT_TYPE_TOGGLE;
				mode = NicoWnnGEvent.Mode.DEFAULT;
				mNicoFirst = false;
				break;

			default:
				break;
		}

		setStatusIcon();
		changeKeyboard(kbd);
		mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.CHANGE_MODE, mode));
		changeKeyboardType(m12keyTable[targetMode]);
	}

	/** @see com.hiroshica.android.input.nicownn2.DefaultSoftKeyboard#initView */
	@Override public View initView(final NicoWnnG parent, final int width, final int height) {
		final View view = super.initView(parent, width, height);
		changeKeyboard(mKeyboard[mCurrentLanguage][mDisplayMode][m12keyTable[mCurrentKeyMode]][mShiftOn][mCurrentKeyMode][0]);
		return view;
	}

	/** @see com.hiroshica.android.input.nicownn2.DefaultSoftKeyboard#changeKeyboard */
	@Override protected boolean changeKeyboard(final MyHeightKeyboard keyboard) {
		resetKeyRepeat();

		if (keyboard != null) {
			if (mIsInputTypeNull) {
				if (mChangeModeKey != null) {
					mChangeModeKey.popupResId = mPopupResId;
				}
			}

			final List<Keyboard.Key> keys = keyboard.getKeys();
			final int keyIndex = (KEY_NUMBER_12KEY < keys.size())
			? KEY_INDEX_CHANGE_MODE_QWERTY : KEY_INDEX_CHANGE_MODE_12KEY;
			mChangeModeKey = keys.get(keyIndex);

			if (mIsInputTypeNull) {
				if (mChangeModeKey != null) {
					mPopupResId = mChangeModeKey.popupResId;
					mChangeModeKey.popupResId = 0;
				}
			}
		}
		return super.changeKeyboard(keyboard);
	}

	/** @see com.hiroshica.android.input.nicownn2.DefaultSoftKeyboard#changeKeyboardType */
	@Override public void changeKeyboardType(final int type) {
		commitText();

		fadePreview();
		resetKeyRepeat();

		m12keyTable[mCurrentKeyMode] = type;
		final MyHeightKeyboard kbd = getTypeChangeKeyboard(type);
		if (kbd != null) {
			/*
			mPrefEditor.putBoolean("opt_enable_qwerty", type == KEYBOARD_QWERTY);
			mPrefEditor.commit();
			 */
			changeKeyboard(kbd);
		}

		setStatusIcon();

		if (m12keyTable[mCurrentKeyMode] == KEYBOARD_12KEY) {
			mWnn.onEvent(mWnn.mEventChangeMode12Key);
		} else {
			mWnn.onEvent(mWnn.mEventChangeModeQwerty);
		}
	}
	/** @see com.hiroshica.android.input.nicownn2.DefaultSoftKeyboard#setPreferences */
	@Override public void setPreferences(final SharedPreferences pref, final EditorInfo editor) {
		super.setPreferences(pref, editor);

		final int inputType = editor.inputType;
		switch (inputType & InputType.TYPE_MASK_CLASS) {

			case InputType.TYPE_CLASS_NUMBER:
			case InputType.TYPE_CLASS_DATETIME:
				mPreferenceKeyMode = KEYMODE_JA_HALF_NUMBER;
				break;

			case InputType.TYPE_CLASS_PHONE:
				if (mHardKeyboardHidden) {
					//mLimitedKeyMode = new int[] {KEYMODE_JA_FULL_NICO};
					mPreferenceKeyMode = KEYMODE_JA_FULL_NICO;
				} else {
					//mLimitedKeyMode = new int[] {KEYMODE_JA_HALF_ALPHABET};
					mPreferenceKeyMode = KEYMODE_JA_FULL_NICO;
				}
				break;

			case InputType.TYPE_CLASS_TEXT:
				switch (inputType & InputType.TYPE_MASK_VARIATION) {

					case InputType.TYPE_TEXT_VARIATION_PASSWORD:
						//mLimitedKeyMode = new int[] {KEYMODE_JA_HALF_ALPHABET, KEYMODE_JA_HALF_NUMBER};
						break;

					case InputType.TYPE_TEXT_VARIATION_EMAIL_ADDRESS:
						//mLimitedKeyMode = new int[] {KEYMODE_JA_HALF_ALPHABET, KEYMODE_JA_HALF_NUMBER};
						break;
					case InputType.TYPE_TEXT_VARIATION_URI:
						//mPreferenceKeyMode = KEYMODE_JA_HALF_ALPHABET;
						mPreferenceKeyMode = KEYMODE_JA_FULL_NICO;
						break;

					default:
						break;
				}
				break;

			default:
				break;
		}

		setPreferencesJAJP(pref, editor);
	}

	/** @see com.hiroshica.android.input.nicownn2.DefaultSoftKeyboard#onUpdateState */
	@Override public void onUpdateState(final NicoWnnG parent) {
		super.onUpdateState(parent);
		if (!mCapsLock) {
			setShiftByEditorInfo();
			if ((true == mNoInput) && (true == mNicoFirst)) {
				resetNicoKeyboard();
			}
		}
	}
	/**
	 * Change the keyboard to default
	 */
	@Override public void setDefaultKeyboard() {
		final Locale locale = Locale.getDefault();
		int keymode = KEYMODE_JA_FULL_NICO;


		if (mPreferenceKeyMode != INVALID_KEYMODE) {
			keymode = mPreferenceKeyMode;
		} else if (mLimitedKeyMode != null) {
			keymode = mLimitedKeyMode[0];
		} else {
			if (!locale.getLanguage().equals(Locale.JAPANESE.getLanguage())) {
				keymode = KEYMODE_JA_HALF_ALPHABET;
			}
		}
		changeKeyMode(keymode);
	}


	/**
	 * Change to the next input mode
	 */
	@Override public void nextKeyMode() {
		/* Search the current mode in the toggle table */
		boolean found = false;
		int index;
		for (index = 0; index < JP_MODE_CYCLE_TABLE.length; index++) {
			if (JP_MODE_CYCLE_TABLE[index] == mCurrentKeyMode) {
				found = true;
				break;
			}
		}

		if (!found) {
			/* If the current mode not exists, set the default mode */
			setDefaultKeyboard();
		} else {
			/* If the current mode exists, set the next input mode */
			final int size = JP_MODE_CYCLE_TABLE.length;
			int keyMode = INVALID_KEYMODE;
			for (int i = 0; i < size; i++) {
				index = (++index) % size;
				if (keyMode == INVALID_KEYMODE) {
					continue;
				}
				if (m12keyTable[keyMode] == KEYBOARD_QWERTY) {
					if (mNoAlphaMode) {
						if (JP_MODE_CYCLE_TABLE[index] == KEYMODE_JA_HALF_ALPHABET) {
							continue;
						}
					}
					if (mNoNumberMode) {
						if (JP_MODE_CYCLE_TABLE[index] == KEYMODE_JA_HALF_NUMBER) {
							continue;
						}
					}
				}
				break;
			}

			if (keyMode != INVALID_KEYMODE) {
				changeKeyMode(keyMode);
			}
		}
	}

	/**
	 * Create the keyboard for portrait mode
	 * <br>
	 * @param parent  The context
	 */
	@Override
	protected void createKeyboardsPortrait() {
		mNicoKeyboard = mSetupKeyboard.SetupSoftKeyboard(mWnn, PORTRAIT, mInputViewHeightIndex, mFlickNicoInput, mFlickGuide, mUse12KeySubTen || isMinimizeSoftKeyboardByHardKeyboard(), mSubTen12KeyMode);

		if (true == isHideSoftKeyboardByHardKeyboard()) {
			return;				// not create soft keyboard
		}

		int[] keyTable = selectQwertyPortKeyTable;
		if (mQwertyMatrixMode) {
			keyTable = selectQwertyPortKeyTable;
		} else {
			keyTable = selectQwertySlantKeyTable;
		}

		MyHeightKeyboard[][] keyList;
		/* qwerty shift_off (portrait) */
		keyList = mKeyboard[LANG_JA][PORTRAIT][KEYBOARD_QWERTY][KEYBOARD_SHIFT_OFF];
		keyList[KEYMODE_JA_FULL_NICO][0]     = new MyHeightKeyboard(mWnn, keyTable[0], mInputViewHeightIndex, KEYTYPE_NICO2, true);

		/* qwerty shift_on (portrait) */
		keyList = mKeyboard[LANG_JA][PORTRAIT][KEYBOARD_QWERTY][KEYBOARD_SHIFT_ON];
		keyList[KEYMODE_JA_FULL_NICO][0]     = new MyHeightKeyboard(mWnn, keyTable[1], mInputViewHeightIndex, KEYTYPE_NICO2, true);

		/* 12-keys shift_off (portrait) */
		keyList = mKeyboard[LANG_JA][PORTRAIT][KEYBOARD_12KEY][KEYBOARD_SHIFT_OFF];

		/* 12-keys shift_on (portrait) */
		keyList = mKeyboard[LANG_JA][PORTRAIT][KEYBOARD_12KEY][KEYBOARD_SHIFT_ON];

		MyHeightKeyboard[][] keyListOn;
		keyList   = mKeyboard[LANG_JA][PORTRAIT][KEYBOARD_12KEY][KEYBOARD_SHIFT_OFF];
		keyListOn = mKeyboard[LANG_JA][PORTRAIT][KEYBOARD_12KEY][KEYBOARD_SHIFT_ON];
		keyList[KEYMODE_JA_FULL_NICO]     = mNicoKeyboard[NICO_MODE_FULL_HIRAGANA][NICO_SLIDE_MODE_TOP];
		keyList[KEYMODE_JA_FULL_NICO_KATAKANA]     = mNicoKeyboard[NICO_MODE_FULL_KATAKANA][NICO_SLIDE_MODE_TOP];
		keyList[KEYMODE_JA_HALF_NICO_KATAKANA]     = mNicoKeyboard[NICO_MODE_HALF_KATAKANA][NICO_SLIDE_MODE_TOP];
		keyListOn[KEYMODE_JA_FULL_NICO]   = mNicoKeyboard[NICO_MODE_FULL_HIRAGANA][NICO_SLIDE_MODE_A];
		keyListOn[KEYMODE_JA_FULL_NICO_KATAKANA]   = mNicoKeyboard[NICO_MODE_FULL_KATAKANA][NICO_SLIDE_MODE_A];
		keyListOn[KEYMODE_JA_HALF_NICO_KATAKANA]   = mNicoKeyboard[NICO_MODE_HALF_KATAKANA][NICO_SLIDE_MODE_A];
		mCreateKeyboard = true;
	}

	/**
	 * Create the keyboard for landscape mode
	 * <br>
	 * @param parent  The context
	 */
	@Override
	protected void createKeyboardsLandscape() {
		mNicoKeyboard = mSetupKeyboard.SetupSoftKeyboard(mWnn, LANDSCAPE, mInputViewHeightIndex, mFlickNicoInput, mFlickGuide, mUse12KeySubTen || isMinimizeSoftKeyboardByHardKeyboard(), mSubTen12KeyMode);

		if (true == isHideSoftKeyboardByHardKeyboard()) {
			return;				// not create soft keyboard
		}

		int[] keyTable = selectQwertyLandKeyTable;
		if (mQwertyMatrixMode) {
			keyTable = selectQwertyLandKeyTable;
		} else {
			keyTable = selectQwertyLandSlantKeyTable;
		}

		MyHeightKeyboard[][] keyList;
		/* qwerty shift_off (landscape) */
		keyList = mKeyboard[LANG_JA][LANDSCAPE][KEYBOARD_QWERTY][KEYBOARD_SHIFT_OFF];
		keyList[KEYMODE_JA_FULL_NICO][0]     = new MyHeightKeyboard(mWnn, keyTable[0], mInputViewHeightIndex, KEYTYPE_NICO2, false);

		/* qwerty shift_on (landscape) */
		keyList = mKeyboard[LANG_JA][LANDSCAPE][KEYBOARD_QWERTY][KEYBOARD_SHIFT_ON];
		keyList[KEYMODE_JA_FULL_NICO][0]     = new MyHeightKeyboard(mWnn, keyTable[1], mInputViewHeightIndex, KEYTYPE_NICO2, false);

		/* 12-keys shift_off (landscape) */
		keyList = mKeyboard[LANG_JA][LANDSCAPE][KEYBOARD_12KEY][KEYBOARD_SHIFT_OFF];

		/* 12-keys shift_on (landscape) */
		keyList = mKeyboard[LANG_JA][LANDSCAPE][KEYBOARD_12KEY][KEYBOARD_SHIFT_ON];

		MyHeightKeyboard[][] keyListOn;
		keyList   = mKeyboard[LANG_JA][LANDSCAPE][KEYBOARD_12KEY][KEYBOARD_SHIFT_OFF];
		keyListOn = mKeyboard[LANG_JA][LANDSCAPE][KEYBOARD_12KEY][KEYBOARD_SHIFT_ON];
		keyList[KEYMODE_JA_FULL_NICO]     = mNicoKeyboard[NICO_MODE_FULL_HIRAGANA][NICO_SLIDE_MODE_TOP];
		keyList[KEYMODE_JA_FULL_NICO_KATAKANA]     = mNicoKeyboard[NICO_MODE_FULL_KATAKANA][NICO_SLIDE_MODE_TOP];
		keyList[KEYMODE_JA_HALF_NICO_KATAKANA]     = mNicoKeyboard[NICO_MODE_HALF_KATAKANA][NICO_SLIDE_MODE_TOP];
		keyListOn[KEYMODE_JA_FULL_NICO]   = mNicoKeyboard[NICO_MODE_FULL_HIRAGANA][NICO_SLIDE_MODE_A];
		keyListOn[KEYMODE_JA_FULL_NICO_KATAKANA]   = mNicoKeyboard[NICO_MODE_FULL_KATAKANA][NICO_SLIDE_MODE_A];
		keyListOn[KEYMODE_JA_HALF_NICO_KATAKANA]   = mNicoKeyboard[NICO_MODE_HALF_KATAKANA][NICO_SLIDE_MODE_A];
		mCreateKeyboard = true;
	}

	/**
	 * Convert the key code to the index of table
	 * <br>
	 * @param index     The key code
	 * @return          The index of the toggle table for input
	 */
	@Override public int getTableIndex(final int keyCode) {
		final int index =
			(keyCode == KEYCODE_JP12_1)     ?  0 :
				(keyCode == KEYCODE_JP12_2)     ?  1 :
					(keyCode == KEYCODE_JP12_3)     ?  2 :
						(keyCode == KEYCODE_JP12_4)     ?  3 :
							(keyCode == KEYCODE_JP12_5)     ?  4 :
								(keyCode == KEYCODE_JP12_6)     ?  5 :
									(keyCode == KEYCODE_JP12_7)     ?  6 :
										(keyCode == KEYCODE_JP12_8)     ?  7 :
											(keyCode == KEYCODE_JP12_9)     ?  8 :
												(keyCode == KEYCODE_JP12_0)     ?  9 :
													(keyCode == KEYCODE_NEWKEY_0)   ? 10 :
														(keyCode == KEYCODE_NEWKEY_1)   ? 11 :
															(keyCode == KEYCODE_NEWKEY_2)   ? 12 :
																(keyCode == KEYCODE_NEWKEY_3)   ? 13 :
																	(keyCode == KEYCODE_NEWKEY_4)   ? 14 :
																		0;
		return index;
	}
	private int checkTableIndex(final int keyCode) {
		final int index =
			(keyCode == KEYCODE_JP12_1)     ?  0 :
				(keyCode == KEYCODE_JP12_2)     ?  1 :
					(keyCode == KEYCODE_JP12_3)     ?  2 :
						(keyCode == KEYCODE_JP12_4)     ?  3 :
							(keyCode == KEYCODE_JP12_5)     ?  4 :
								(keyCode == KEYCODE_JP12_6)     ?  5 :
									(keyCode == KEYCODE_JP12_7)     ?  6 :
										(keyCode == KEYCODE_JP12_8)     ?  7 :
											(keyCode == KEYCODE_JP12_9)     ?  8 :
												(keyCode == KEYCODE_JP12_0)     ?  9 :
													-1;
		return index;
	}

	/**
	 * Get the toggle table for input that is appropriate in current mode.
	 *
	 * @return      The toggle table for input
	 */
	private String[][] getCycleTable() {
		final String[][] cycleTable = mCycleTable[mNicoKeyMode][mChangeAlphaBigMode];
		return cycleTable;
	}

	/**
	 * Get the replace table that is appropriate in current mode.
	 *
	 * @return      The replace table
	 */
	private HashMap<String,String> getReplaceTable() {
		HashMap<String,String> hashTable = null;
		switch (mCurrentKeyMode) {
			case KEYMODE_JA_FULL_NICO:
				hashTable = mSetupKeyboard.SetupReplaceTable(mTsuMode);
				break;

			default:
				break;
		}
		return hashTable;
	}

	/**
	 * Set the status icon that is appropriate in current mode
	 */
	@Override
	public void setStatusIcon() {
		int icon = 0;

		switch (mCurrentKeyMode) {
			case KEYMODE_JA_FULL_NICO:
				icon = mSetupKeyboard.SetupIcon();
				break;

			default:
				break;
		}

		mWnn.showStatusIcon(icon);
	}

	/** @see com.hiroshica.android.input.nicownn2.DefaultSoftKeyboard#setHardKeyboardHidden */
	@Override public void setHardKeyboardHidden(final boolean hidden, final boolean hardkey) {
		if (mWnn != null) {
			if (!hidden) {
				mWnn.onEvent(mWnn.mEventChangeModeQwerty);
			}

			if (mHardKeyboardHidden != hidden) {
				if ((mLimitedKeyMode != null) || ((mCurrentKeyMode != KEYMODE_JA_FULL_NICO))) {

					mLastInputType = InputType.TYPE_NULL;
					if (mWnn.isInputViewShown()) {
						setDefaultKeyboard();
					}
				}
			}
		}
		super.setHardKeyboardHidden(hidden, hardkey);
	}

	/**
	 * Change the key-mode to the allowed one which is restricted
	 *  by the text input field or the type of the keyboard.
	 * @param keyMode The key-mode
	 * @return the key-mode allowed
	 */
	private int filterKeyMode(final int keyMode) {
		int targetMode = keyMode;
		final int[] limits = mLimitedKeyMode;
		/* restrict by the type of the text field */
		if (limits != null) {
			boolean hasAccepted = false;
			boolean hasRequiredChange = true;
			final int size = limits.length;
			final int nowMode = mCurrentKeyMode;

			for (int i = 0; i < size; i++) {
				if (targetMode == limits[i]) {
					hasAccepted = true;
					break;
				}
				if (nowMode == limits[i]) {
					hasRequiredChange = false;
				}
			}

			if (!hasAccepted) {
				if (hasRequiredChange) {
					targetMode = mLimitedKeyMode[0];
				} else {
					targetMode = INVALID_KEYMODE;
				}
			}
		}

		return targetMode;
	}

	@Override public String[] convertFlickToKeyString(int flickdir) {
		final int col = getTableIndex(mPrevInputKeyCode);
		final int row = flickdir;
		final String[][] cycleTable = getCycleTable();

		if (null != cycleTable) {
			return cycleTable[col * 15 + row];
		}
		return null;
	}

	/*****************************************
	 * onkey (normal)
	 */
	/** @see com.hiroshica.android.input.nicownn2.DefaultSoftKeyboard#onKey */
	@Override public void onKey(final int primaryCode2, final int[] keyCodes) {
		int primaryCode = primaryCode2;
		boolean changeShiftLock = true;

		MyHeightKeyboard newKeyboard = null;
		if (mDisableKeyInput) {
			return;
		}
		final int getFlickCode = checkFlickKeyCode(false, false);
		if (getFlickCode >= 0) {
			// get flick keydata
			final String keyString[] = convertFlickToKeyString(getFlickCode);
			if (null != keyString) {
				mWnn.onEvent(mWnn.mEventTouchOtherKey);
				mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.TOGGLE_CHAR, keyString));
			}
			mNicoFirst = false;
			mNicoFlick = false;
			return;
		} else if (getFlickCode < -1) {
			switch (primaryCode) {
			case KEYCODE_JP12_CONVPREDICT:
			case KEYCODE_JP12_CONVPREDICT_BACKWARD:
				break;
			default:
				primaryCode = getFlickCode;
			}
		}

		switch (primaryCode) {
			case KEYCODE_JP12_TOGGLE_MODE:
			case KEYCODE_JP12_TOGGLE_MODE2:
			case KEYCODE_QWERTY_TOGGLE_MODE:
			case KEYCODE_QWERTY_TOGGLE_MODE2:
				if (!mNoInput) {
					final NicoWnnGJAJP wnn = NicoWnnGJAJP.getInstance();
					wnn.changeAlphaKanaDirectPhase();
					break;
				}
				if (!mIsInputTypeNull) {
					nextKeyMode();
				}
				break;

			case DefaultSoftKeyboard.KEYCODE_QWERTY_BACKSPACE:
			case KEYCODE_JP12_BACKSPACE:
				if (mNicoFirst == true) {
					resetNicoKeyboard();
				}
				else {
					mWnn.onEvent(mWnn.mEventInputKeyDel);
				}
				changeShiftLock = false;
				break;

			case DefaultSoftKeyboard.KEYCODE_QWERTY_SHIFT:
				if (mKeyRepeatCount == 2) {
					toggleShiftLock(3);
					changeShiftLock = false;
					break;
				} else if (mKeyRepeatCount == 0) {
					changeShiftLock = false;
					break;
				}
				toggleShiftLock(1);
				changeShiftLock = false;
				break;

			case DefaultSoftKeyboard.KEYCODE_QWERTY_ALT:
				processAltKey();
				break;

			case KEYCODE_QWERTY_ENTER:
			case KEYCODE_JP12_ENTER:
				if (mNicoFirst == false) {
					mWnn.onEvent(mWnn.mEventInputEnter);
				}
				break;

			case KEYCODE_QWERTY_KBD:
			case KEYCODE_JP12_KBD:
				resetNicoKeyboard();
				break;

			case KEYCODE_JP12_EMOJI:
			case KEYCODE_QWERTY_EMOJI:
				if (mNicoFirst == false) {
					commitText();
					mWnn.onEvent(mWnn.mEventChangeModeSymbol);
				}
				break;

			case KEYCODE_EISU_KANA:
				if (mNicoFirst == false) {
					mWnn.onEvent(mWnn.mEventChangeModeEisuKana);
				}
				break;

			case KEYCODE_JP12_SHARP:
				if ((mNicoFirst == false) && mNoInput) {
					commitText();
					mWnn.onEvent(mWnn.mEventChangeModeDocomo);
				}
				break;

			case KEYCODE_JP12_1:
			case KEYCODE_JP12_2:
			case KEYCODE_JP12_3:
			case KEYCODE_JP12_4:
			case KEYCODE_JP12_5:
			case KEYCODE_JP12_6:
			case KEYCODE_JP12_7:
			case KEYCODE_JP12_8:
			case KEYCODE_JP12_9:
			case KEYCODE_JP12_0:
			case KEYCODE_NEWKEY_0:
			case KEYCODE_NEWKEY_1:
			case KEYCODE_NEWKEY_2:
			case KEYCODE_NEWKEY_3:
			case KEYCODE_NEWKEY_4:
				if (mNicoFirst == false) {
					mWnn.onEvent(mWnn.mEventTouchOtherKey);
					if (NICOFLICK_NONE == mFlickNicoInput) {
						mPrevInputKeyCode = primaryCode;
					}
					else{
						if (-1 == checkTableIndex(mPrevInputKeyCode)) {
							mPrevInputKeyCode = primaryCode;
						}
					}
					mNicoFirst = true;

					/* change keymap */
					final int index = getTableIndex(mPrevInputKeyCode);
					final MyHeightKeyboard[][] keyList = mKeyboard[LANG_JA][mDisplayMode][KEYBOARD_12KEY][KEYBOARD_SHIFT_ON];

					keyList[KEYMODE_JA_FULL_NICO] = mNicoKeyboard[mNicoKeyMode][index+1];
					newKeyboard = getShiftChangeKeyboard(KEYBOARD_SHIFT_ON);
					mShiftOn = KEYBOARD_SHIFT_ON;
					changeKeyboard(newKeyboard);
				}
				else{
					/* change keyboard */
					newKeyboard = getShiftChangeKeyboard(KEYBOARD_SHIFT_OFF);
					mShiftOn = KEYBOARD_SHIFT_OFF;
					changeKeyboard(newKeyboard);
					/* set input char & update text */
					final int col = getTableIndex(mPrevInputKeyCode);
					final int row = getTableIndex(primaryCode);
					final String[][] cycleTable = getCycleTable();
					mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.TOGGLE_CHAR, cycleTable[col * 15 + row]));
					mNicoFirst = false;
					mNicoFlick = false;
				}
				break;

			case KEYCODE_JP12_ASTER:
				if (mNicoFirst == true) {
					resetNicoKeyboard();
				} else {
					if (!mNoInput) {
						/* Processing to toggle Dakuten, Handakuten, and capital */
						final HashMap<String,String> replaceTable = getReplaceTable();
						if (replaceTable == null) {
							Log.e("NicoWnnG", "not founds replace table");
						} else {
							mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.REPLACE_CHAR, replaceTable));
							mPrevInputKeyCode = primaryCode;
						}
					} else {
						mWnn.onEvent(mWnn.mEventInputEnter);
					}
				}
				break;

			case KEYCODE_JP12_SPACE:
				if (mNicoFirst == false) {
					NicoWnnGJAJP t = NicoWnnGJAJP.getInstance();
					if (t.candidatesViewManagerIsShown() && t.candidatesViewManagerIsIndicated()) {
						mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_SOFT_KEY, new KeyEvent(KeyEvent.ACTION_DOWN, KeyEvent.KEYCODE_SPACE)));
					} else if (t.getInstance().isRenbun()||t.isPredict()) {
						mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_KEY, new KeyEvent(KeyEvent.ACTION_DOWN, KeyEvent.KEYCODE_SPACE)));
						mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_CHAR, ' '));
					} else if ((mCurrentKeyMode == KEYMODE_JA_FULL_HIRAGANA) && !mNoInput) {
						mWnn.onEvent(mWnn.mEventConvert);
					} else if ((mCurrentKeyMode == KEYMODE_JA_FULL_NICO) && !mNoInput) {
						mWnn.onEvent(mWnn.mEventConvert);
					} else if ((mCurrentKeyMode == KEYMODE_JA_FULL_NICO_KATAKANA) && !mNoInput) {
						mWnn.onEvent(mWnn.mEventConvert);
					} else if (t.candidatesViewManagerIsShown()) {
						mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_SOFT_KEY, new KeyEvent(KeyEvent.ACTION_DOWN, KeyEvent.KEYCODE_SPACE)));
					} else if ((mCurrentKeyMode == KEYMODE_JA_FULL_HIRAGANA) && mNoInput && mKana12SpaceZen) {
						mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_CHAR, '\u3000'));
					} else if ((mCurrentKeyMode == KEYMODE_JA_FULL_NICO) && mNoInput && mKana12SpaceZen) {
						mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_CHAR, '\u3000'));
					} else if ((mCurrentKeyMode == KEYMODE_JA_FULL_NICO_KATAKANA) && mNoInput && mKana12SpaceZen) {
						mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_CHAR, '\u3000'));
					} else if ((mCurrentKeyMode == KEYMODE_JA_FULL_ALPHABET) && mNoInput && mKana12SpaceZen) {
						mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_CHAR, '\u3000'));
					} else if ((mCurrentKeyMode == KEYMODE_JA_FULL_NUMBER) && mNoInput && mKana12SpaceZen) {
						mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_CHAR, '\u3000'));
					} else {
						mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_CHAR, ' '));
					}
				}
				break;

			case KEYCODE_JP12_ZEN_SPACE:
				if (mNicoFirst == false) {
					NicoWnnGJAJP t = NicoWnnGJAJP.getInstance();
					if (t.candidatesViewManagerIsShown() && t.candidatesViewManagerIsIndicated()) {
						mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_SOFT_KEY, new KeyEvent(KeyEvent.ACTION_DOWN, KeyEvent.KEYCODE_SPACE)));
					} else if (t.getInstance().isRenbun()||t.isPredict()) {
						mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_KEY, new KeyEvent(KeyEvent.ACTION_DOWN, KeyEvent.KEYCODE_SPACE)));
					} else if ((mCurrentKeyMode == KEYMODE_JA_FULL_HIRAGANA) && !mNoInput) {
						mWnn.onEvent(mWnn.mEventConvert);
					} else if ((mCurrentKeyMode == KEYMODE_JA_FULL_NICO) && !mNoInput) {
						mWnn.onEvent(mWnn.mEventConvert);
					} else if ((mCurrentKeyMode == KEYMODE_JA_FULL_NICO_KATAKANA) && !mNoInput) {
						mWnn.onEvent(mWnn.mEventConvert);
					} else if (t.candidatesViewManagerIsShown()) {
						mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_SOFT_KEY, new KeyEvent(KeyEvent.ACTION_DOWN, KeyEvent.KEYCODE_SPACE)));
					} else {
						mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_CHAR, '\u3000'));
					}
				}
				break;

			case KEYCODE_SWITCH_FULL_NICO:
				/* Change mode to Full width nicotouch */
				changeKeyMode(KEYMODE_JA_FULL_NICO);
				break;

			case KEYCODE_SELECT_CASE:
				final int shifted = (mShiftOn == 0) ? 1 : 0;
				newKeyboard = getShiftChangeKeyboard(shifted);
				if (newKeyboard != null) {
					mShiftOn = shifted;
					changeKeyboard(newKeyboard);
				}
				break;

			case KEYCODE_JP12_CLOSE:
				mWnn.onEvent(mWnn.mEventInputBack);
				break;

			case KEYCODE_JP12_LEFT:
				mWnn.onEvent(mWnn.mEventInputDpadLeft);
				changeShiftLock = false;
				break;

			case KEYCODE_JP12_RIGHT:
				mWnn.onEvent(mWnn.mEventInputDpadRight);
				changeShiftLock = false;
				break;

			case KEYCODE_JP12_UP:
				mWnn.onEvent(mWnn.mEventInputDpadUp);
				changeShiftLock = false;
				break;

			case KEYCODE_JP12_DOWN:
				mWnn.onEvent(mWnn.mEventInputDpadDown);
				changeShiftLock = false;
				break;

			case KEYCODE_NOP:
				break;

			case KEYCODE_PREFERENCE_SETTING:
				if (mWnn != null) {
					try {
						mWnn.openPreferenceSetting();
					} catch (final Exception e) {
						//
					}
				}
				break;

			case KEYCODE_SHOW_HELP:
				if (mWnn != null) {
					//try {
					mWnn.openHelp();
					//} catch (final Exception e) {
					//
					//}
				}
				break;

			case KEYCODE_MUSHROOM:
				if (mWnn != null) {
					final String str = mWnn.getComposingText(ComposingText.LAYER2);
					mWnn.onEvent(mWnn.mEventInputBack);
					mWnn.invokeMushroom(str);
				}
				break;

			case KEYCODE_JP12_CONVPREDICT:
				mWnn.onEvent(mWnn.mEventConvertPredict);
				break;

			case KEYCODE_JP12_CONVPREDICT_BACKWARD:
				mWnn.onEvent(mWnn.mEventConvertPredictBackward);
				break;

			case KEYCODE_USERSYMBOL_ZEN_HIRAGANA:
			case KEYCODE_USERSYMBOL_ZEN_KATAKANA:
			case KEYCODE_USERSYMBOL_ZEN_ALPHABET:
			case KEYCODE_USERSYMBOL_ZEN_NUMBER:
			case KEYCODE_USERSYMBOL_HAN_KATAKANA:
			case KEYCODE_USERSYMBOL_HAN_ALPHABET:
			case KEYCODE_USERSYMBOL_HAN_NUMBER:
				commitText();
				mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.CHANGE_MODE, NicoWnnGJAJP.ENGINE_MODE_USERSYMBOL, -primaryCode));
				break;

			default:
				if (primaryCode >= 0) {
					if (mKeyboardView.isShifted()) {
						primaryCode = Character.toUpperCase(primaryCode);
					}
					mWnn.onEvent(new NicoWnnGEvent(NicoWnnGEvent.INPUT_CHAR, (char)primaryCode));
				}
				break;
		}

		if (changeShiftLock) {
			toggleShiftLock(2);
		}

		/* update shift key's state */
		if (!mCapsLock && (primaryCode != DefaultSoftKeyboard.KEYCODE_QWERTY_SHIFT)) {
			setShiftByEditorInfo();
		}
	}

	/*
	 * reset nicotouch keyboard
	 */
	@Override public void resetNicoKeyboard() {
		final MyHeightKeyboard newKeyboard = getShiftChangeKeyboard(KEYBOARD_SHIFT_OFF);
		if (newKeyboard != null) {
			mShiftOn = KEYBOARD_SHIFT_OFF;
			changeKeyboard(newKeyboard);
		}
		mNicoFirst = false;
		mNicoFlick = false;
	} // resetNicoKeyboard

	/**
	 */
	@Override public void onPress(final int primaryCode) {
		super.onPress(primaryCode);
	}
	/**
	 */
	@Override public void onRelease(final int primaryCode) {
		super.onRelease(primaryCode);
	} // onRelease
	/**
	 */
	@Override public int convertModeFlick(final int prev, final int key) {
		return mSetupKeyboard.GetFlickKeyCode(prev, key);
	}



	private static int selectQwertyLandKeyTable[] = {
		R.xml.keyboard_qwerty_jp_0,
		R.xml.keyboard_qwerty_jp_shift_0,
	};

	private static int selectQwertyPortKeyTable[] = {
		R.xml.keyboard_qwerty_jp_0,
		R.xml.keyboard_qwerty_jp_shift_0,
	};

	private static int selectQwertySlantKeyTable[] = {
		R.xml.keyboard_qwerty_jp_s0,
		R.xml.keyboard_qwerty_jp_shift_s0,
	};

	private static int selectQwertyLandSlantKeyTable[] = {
		R.xml.keyboard_qwerty_jp_s0,
		R.xml.keyboard_qwerty_jp_shift_s0,
	};


}
